#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "bitmap.h"
#include "prime.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

int main (int argc, char *argv[])
{
	prime_table_t ptab;
	double begin;
	uint64_t *bm;
	int threads;
	FILE *fp;
	fpos_t pos;
	char buf[256];
	long *nlist, *rlist, nlistlen, nlistsize, maxn;
	long n, k, k0, d, m, s;
	
	if ( argc < 2 ) { puts ("greedyverify filename [threads]"); return 0; }
	fp = fopen (argv[1],"r");
	if ( ! fp ) { printf ("Unable to open file %s\n", argv[1]); return 0; }
	if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 1 of file %s\n", argv[1]); return 0; }
	if ( sscanf (buf, "k0=%ld", &k0) != 1 )  { printf ("Error parsing line 1 of file %s, expected k0=%%ld\n", argv[1]); return 0; }
	if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 2 of file %s\n", argv[1]); return 0; }
	if ( sscanf (buf, "s=%ld", &s) != 1 )  { printf ("Error parsing line 2 of file %s, expected s=%%ld\n", argv[1]); return 0; }
	if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 3 of file %s\n", argv[1]); return 0; }
	if ( sscanf (buf, "d=%ld", &d) != 1 )  { printf ("Error parsing line 3 of file %s, expected d=%%ld\n", argv[1]); return 0; }
	if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 4 of file %s\n", argv[1]); return 0; }
	if ( sscanf (buf, "m=%ld", &m) != 1 && sscanf (buf, "n=%ld", &m) != 1 )  { printf ("Error parsing line 4 of file %s, expected m=%%ld\n", argv[1]); return 0; }
	if ( fgetpos (fp, &pos) ) { printf ("fgetpost failed on file %s\n", argv[1]); return 0; }
	for ( nlistsize = 0 ; fgets(buf, sizeof(buf), fp) ; nlistsize++ ) if ( sscanf (buf, "%ld-%ld\n", &k, &n) == 2 ) nlistsize += n-k;
	if ( fsetpos (fp, &pos) ) { printf ("fsetpost failed on file %s\n", argv[1]); return 0; }	
	nlist = malloc ((nlistsize+1)*sizeof(*nlist));
	rlist = malloc ((nlistsize+1)*sizeof(*rlist));
	maxn = m;
	for ( nlistlen = 0 ; fgets(buf, sizeof(buf), fp) ; nlistlen++ ) {
		if ( sscanf (buf, "%ld,%ld\n", nlist+nlistlen, rlist+nlistlen) != 2 ) {
			if ( sscanf (buf, "%ld-%ld\n", nlist+nlistlen, &n) == 2 && n > nlist[nlistlen] ) {
				rlist[nlistlen] = 0;
				for ( long i = nlist[nlistlen]+1 ; i <= n ; i++ ) { nlist[++nlistlen] = i; rlist[nlistlen] = 0; }
			} else if ( sscanf (buf, "%ld\n", nlist+nlistlen) == 1 ) {
				rlist[nlistlen] = 0;
			} else {
				printf ("Error parsing line %s in file %s\n", buf, argv[1]);
				return 0;
			}
		}
		if ( nlist[nlistlen] > maxn ) maxn = nlist[nlistlen];
	}
	nlist[nlistlen] = 0;
	assert (nlistlen == nlistsize);
	fclose (fp);
	if ( argc > 2 ) threads = atoi(argv[2]); else threads = 0;
	
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);
	
	prime_table_init (ptab, 2, prime_pi(k0)+1);	// we need to go one prime past k0
	
	printf ("k0=%ld, s=%ld, d=%ld, p_%ld=%ld (and %ld of the next %ld primes)\n", k0, s, d, m, prime_table_nth_prime(ptab, m), nlistlen, maxn-m);

	begin = omp_get_wtime();
	bm = survivor_bitmap (s, d+1, m, nlist, rlist, ptab, threads);
	k = bm_weight (bm, d+1);
	printf ("greedy sieve  of [%ld,%ld] using Schinzel bound p_%ld = %ld and sieving %ld of the next %ld primes yielded %ld survivors (%.1fs)\n", s, s+d, m, prime_table_nth_prime (ptab, m), nlistlen, maxn-m, k, omp_get_wtime()-begin);
	if ( k < k0 ) { printf ("Not enough survivors, you need to increase the diameter or decrease the prime bound\n"); return 0; }	
	
	#pragma omp parallel num_threads(threads)
	{
		double start;
		uint64_t *rm;
		unsigned char *pgap;
		int tid;
		register long i, j, p, w;

		tid = omp_get_thread_num();
		start = omp_get_wtime();
		rm = bm_alloc (((k0>>6)+1)<<6);
		p = prime_table_nth_gap (ptab, &pgap, m);
		for ( int  i = 0 ; p <= k0 && i < tid ; i++ ) p += 2*(*pgap++);
		w = ((p>>6)+1)<<6;  j = 0;
		while ( p <= k0 ) {
			// don't bother checking primes that we sieved -- this check assumes nlist is sorted, but it won't break anything if it isn't
			while ( nlist[j] && nlist[j] < prime_table_gap_pi (ptab, pgap) ) j++;
			if ( prime_table_gap_pi (ptab, pgap) != nlist[j] ) { 
				if ( w > (p>>2) ) {
					bm_clear (rm, p);
					bm_residue_set (rm, p, bm, d+1);
					i = bm_min_unset (rm, p);
					w = ((i>>6)+1)<<6;
					if ( i == p ) { printf ("Inadmissible modulo p_%ld=%ld, terminating.\n", prime_table_pi (ptab, p), p); exit(-1); }
				} else {
					bm_clear (rm, w);
					bm_partial_residue_set (rm, w>>6, p, bm, d+1);
					i = bm_min_unset (rm, w);
					if ( i == w ) { w*=2; if ( w > p ) w = ((p>>6)+1)<<6; continue; }
					w = ((i>>6)+1)<<6;
				}
			}
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("verified to p_%ld = %ld (%.1fs)\n", prime_table_pi (ptab, p), p, end-begin); start = end; }
			}
			for ( int i = 0 ; p <= k0 && i < threads ; i++ ) p += 2*(*pgap++);
		}
		bm_free (rm);
	}
	free (nlist);  free(rlist);
	printf ("verified to p_%ld = %ld (%.1fs)\n", prime_table_pi (ptab, k0), prime_table_nth_prime (ptab, prime_table_pi (ptab, k0)), omp_get_wtime()-begin);
	d = survivor_bitmap_best_subsequence (&n, bm, d+1, k0);
	assert (d);
	s += n;
	printf ("Verified admissible sequence in [%ld,%ld] with k=%ld entries using Schinzel bound P_%ld=%ld and sieving %ld of the next %ld primes, obtaining diameter d=%ld\n", s, s+d, k0, m, prime_table_nth_prime (ptab, m), nlistlen, maxn-m, d);
	bm_free (bm);
	prime_table_clear (ptab);
	prime_cleanup();

	printf ("Total time %.1fs\n", omp_get_wtime()-begin);
}
