#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "prime.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

int main (int argc, char *argv[])
{
	prime_table_t tab;
	double begin;
	int threads;
	unsigned char *gap, *egap;
	long k0, m, p, q, d, e;
	
	if ( argc < 2 ) { puts ("zhangsearch k0 [threads]"); return 0; }
	k0 = atol(argv[1]);
	assert ( k0 > 2 );
	if ( argc > 2 ) threads = atoi(argv[2]); else threads = omp_get_max_threads();
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);

	begin = omp_get_wtime();
	m = prime_pi (k0);
	prime_table_init (tab, 2, m+k0);
	printf ("Built prime table from p_%ld=%ld to p_%ld=%ld (%.1fs)\n", tab->si, tab->sp, tab->ti, tab->tp, omp_get_wtime()-begin);
	printf ("First k primes past k yields diameter %ld\n", prime_table_nth_prime (tab, m+k0)-prime_table_nth_prime (tab, m+1));
	m = zhang_best_m (tab, k0, begin, threads);
	
	// now check whether we can actually get a better diameter for some larger m < pi(k0)
	q = prime_table_nth_prime (tab, m+k0);
	d = e = q-prime_table_nth_prime (tab, m+1);
	gap = tab->gaps + m+1 - tab->si;
	egap = tab->gaps + prime_table_pi(tab, k0)+1 - tab->si;
	for ( ; gap < egap ; gap++ ) {
		e += 2*((long)*(gap+k0-1) - (long)*gap);
		if ( e < d ) { d = e;  m = gap - tab->gaps + tab->si; printf ("m=%ld gives diameter %ld\n", m, d);}
	}
	p = prime_table_nth_prime (tab, m+1);  q = prime_table_nth_prime (tab, m+k0);
	prime_table_clear (tab);
	prime_cleanup();
	printf ("Optimal m = %ld gives diameter %ld for interval from p_%ld=%ld to p_%ld=%ld\n", m, q-p, m+1, p, m+k0, q);
	printf ("This result is only heuristic, you now need to run\n    zhangverify %ld %ld\n", k0, m);
	printf ("Total elapsed time %.1fs\n", omp_get_wtime()-begin);
}
